@@ -61,7 +61,10 @@ use std::collections::BTreeMap;
 use std::fs;
 use std::io::{self, prelude::*, BufReader, BufWriter, Result};
 use std::path::{Path, PathBuf};
-use std::sync::mpsc;
+use std::sync::{
+    atomic::{AtomicU64, Ordering},
+    mpsc, Arc,
+};
 
 const SSTABLE_DIR: &str = "sstables";
 
@@ -112,6 +115,22 @@ impl Default for Config {
     }
 }
 
+struct WorkerStats {
+    read_bytes: AtomicU64,
+    written_bytes: AtomicU64,
+}
+
+#[derive(Debug, Clone, Copy)]
+pub struct Stats {
+    resident_bytes: u64,
+    on_disk_bytes: u64,
+    logged_bytes: u64,
+    written_bytes: u64,
+    read_bytes: u64,
+    space_amp: f64,
+    write_amp: f64,
+}
+
 fn hash<const K: usize, const V: usize>(k: &[u8; K], v: &Option<[u8; V]>) -> u32 {
     let mut hasher = crc32fast::Hasher::new();
     hasher.update(&[v.is_some() as u8]);
@@ -140,69 +159,57 @@ struct Worker<const K: usize, const V: usize> {
     db_sz: u64,
     path: PathBuf,
     config: Config,
+    stats: Arc<WorkerStats>,
 }
 
 impl<const K: usize, const V: usize> Worker<K, V> {
     fn run(mut self) {
-        loop {
-            match self.inbox.try_recv() {
-                Ok(WorkerMessage::NewSST { id, sst_sz, db_sz }) => {
-                    self.db_sz = db_sz;
-                    self.sstable_directory.insert(id, sst_sz);
+        let mut running = true;
+        while running {
+            match self.inbox.recv() {
+                Ok(message) => {
+                    running = self.handle_message(message);
                 }
-                Ok(WorkerMessage::Stop(dropper)) => {
-                    drop(dropper);
-                    return;
-                }
-                Ok(WorkerMessage::Heartbeat(dropper)) => {
-                    drop(dropper);
-                }
-                Err(mpsc::TryRecvError::Empty) => {}
-                Err(mpsc::TryRecvError::Disconnected) => {
-                    log::info!("tiny-lsm compaction worker quitting");
-                    return;
-                }
-            };
-
-            let res = self.sstable_maintenance();
-            match res {
-                Ok(false) => {}
-                Ok(true) => continue,
-                Err(e) => {
-                    log::error!("error while compacting sstables in the background: {:?}", e);
+                Err(mpsc::RecvError) => {
+                    break;
                 }
             }
 
-            // no work to do, wait for new sstables
-            match self.inbox.recv() {
-                Ok(WorkerMessage::NewSST { id, sst_sz, db_sz }) => {
-                    self.db_sz = db_sz;
-                    self.sstable_directory.insert(id, sst_sz);
-                }
-                Ok(WorkerMessage::Stop(dropper)) => {
-                    drop(dropper);
-                    log::info!("tiny-lsm compaction worker quitting");
-                    return;
-                }
-                Ok(WorkerMessage::Heartbeat(dropper)) => {
-                    drop(dropper);
-                }
-                Err(mpsc::RecvError) => {
-                    log::info!("tiny-lsm compaction worker quitting");
-                    return;
-                }
+            // only compact one run at a time before checking
+            // for new messages.
+            if let Err(e) = self.sstable_maintenance() {
+                log::error!("error while compacting sstables in the background: {:?}", e);
+            }
+        }
+        log::info!("tiny-lsm compaction worker quitting");
+    }
+
+    fn handle_message(&mut self, message: WorkerMessage) -> bool {
+        match message {
+            WorkerMessage::NewSST { id, sst_sz, db_sz } => {
+                self.db_sz = db_sz;
+                self.sstable_directory.insert(id, sst_sz);
+                true
+            }
+            WorkerMessage::Stop(dropper) => {
+                drop(dropper);
+                false
+            }
+            WorkerMessage::Heartbeat(dropper) => {
+                drop(dropper);
+                true
             }
         }
     }
 
-    fn sstable_maintenance(&mut self) -> Result<bool> {
+    fn sstable_maintenance(&mut self) -> Result<()> {
         let on_disk_size: u64 = self.sstable_directory.values().sum();
 
         log::debug!("disk size: {} mem size: {}", on_disk_size, self.db_sz);
         if self.sstable_directory.len() > 1
             && on_disk_size / (self.db_sz + 1) > self.config.max_space_amp
         {
-            log::info!(
+            log::debug!(
                 "performing full compaction, decompressed on-disk \
                 database size has grown beyond {}x the in-memory size",
                 self.config.max_space_amp
@@ -210,11 +217,11 @@ impl<const K: usize, const V: usize> Worker<K, V> {
             let run_to_compact: Vec<u64> = self.sstable_directory.keys().copied().collect();
 
             self.compact_sstable_run(&run_to_compact)?;
-            return Ok(true);
+            return Ok(());
         }
 
         if self.sstable_directory.len() < self.config.merge_window {
-            return Ok(false);
+            return Ok(());
         }
 
         for window in self
@@ -231,13 +238,17 @@ impl<const K: usize, const V: usize> Worker<K, V> {
                 let run_to_compact: Vec<u64> = window.into_iter().map(|(id, _sum)| **id).collect();
 
                 self.compact_sstable_run(&run_to_compact)?;
-                return Ok(true);
+                return Ok(());
             }
         }
 
-        Ok(false)
+        Ok(())
     }
 
+    // This function must be able to crash at any point without
+    // leaving the system in an unrecoverable state, or without
+    // losing data. This function must be nullipotent from the
+    // external API surface's perspective.
     fn compact_sstable_run(&mut self, sstable_ids: &[u64]) -> Result<()> {
         log::debug!(
             "trying to compact sstable_ids {:?}",
@@ -249,12 +260,19 @@ impl<const K: usize, const V: usize> Worker<K, V> {
 
         let mut map = BTreeMap::new();
 
+        let mut read_pairs = 0;
+
         for sstable_id in sstable_ids {
             for (k, v) in read_sstable::<K, V>(&self.path, *sstable_id)? {
                 map.insert(k, v);
+                read_pairs += 1;
             }
         }
 
+        self.stats
+            .read_bytes
+            .fetch_add(read_pairs * (4 + 1 + K + V) as u64, Ordering::Relaxed);
+
         let sst_id = sstable_ids
             .iter()
             .max()
@@ -262,6 +280,10 @@ impl<const K: usize, const V: usize> Worker<K, V> {
 
         write_sstable(&self.path, *sst_id, &map, true, &self.config)?;
 
+        self.stats
+            .written_bytes
+            .fetch_add(map.len() as u64 * (4 + 1 + K + V) as u64, Ordering::Relaxed);
+
         let sst_sz = map.len() as u64 * (4 + K + V) as u64;
         self.sstable_directory.insert(*sst_id, sst_sz);
 
@@ -276,7 +298,6 @@ impl<const K: usize, const V: usize> Worker<K, V> {
                 .remove(sstable_id)
                 .expect("compacted sst not present in sstable_directory");
         }
-
         fs::File::open(self.path.join(SSTABLE_DIR))?.sync_all()?;
 
         Ok(())
@@ -287,7 +308,7 @@ fn id_format(id: u64) -> String {
     format!("{:016x}", id)
 }
 
-fn list_sstables(path: &Path) -> Result<BTreeMap<u64, u64>> {
+fn list_sstables(path: &Path, remove_tmp: bool) -> Result<BTreeMap<u64, u64>> {
     let mut sstable_map = BTreeMap::new();
 
     for dir_entry_res in fs::read_dir(path.join(SSTABLE_DIR))? {
@@ -303,7 +324,7 @@ fn list_sstables(path: &Path) -> Result<BTreeMap<u64, u64>> {
 
             sstable_map.insert(id, metadata.len());
         } else {
-            if file_name.ends_with("-tmp") {
+            if remove_tmp && file_name.ends_with("-tmp") {
                 log::warn!("removing incomplete sstable rewrite {}", file_name);
                 fs::remove_file(path.join(SSTABLE_DIR).join(file_name))?;
             }
@@ -337,18 +358,18 @@ fn write_sstable<const K: usize, const V: usize>(
             .expect("zstd encoder failure"),
     );
 
-    bw.write(&(items.len() as u64).to_le_bytes())?;
+    bw.write_all(&(items.len() as u64).to_le_bytes())?;
 
     for (k, v) in items {
         let crc: u32 = hash(k, v);
-        bw.write(&crc.to_le_bytes())?;
-        bw.write(&[v.is_some() as u8])?;
-        bw.write(k)?;
+        bw.write_all(&crc.to_le_bytes())?;
+        bw.write_all(&[v.is_some() as u8])?;
+        bw.write_all(k)?;
 
         if let Some(v) = v {
-            bw.write(v)?;
+            bw.write_all(v)?;
         } else {
-            bw.write(&[0; V])?;
+            bw.write_all(&[0; V])?;
         }
     }
 
@@ -424,22 +445,20 @@ fn read_sstable<const K: usize, const V: usize>(
 
 pub struct Lsm<const K: usize, const V: usize> {
     // `BufWriter` flushes on drop
-    log: BufWriter<fs::File>,
     memtable: BTreeMap<[u8; K], Option<[u8; V]>>,
     db: BTreeMap<[u8; K], [u8; V]>,
     worker_outbox: mpsc::Sender<WorkerMessage>,
     next_sstable_id: u64,
     dirty_bytes: usize,
+    log: BufWriter<fs::File>,
     path: PathBuf,
     config: Config,
+    stats: Stats,
+    worker_stats: Arc<WorkerStats>,
 }
 
 impl<const K: usize, const V: usize> Drop for Lsm<K, V> {
     fn drop(&mut self) {
-        if let Err(e) = self.flush() {
-            log::error!("failed to flush while dropping Lsm: {:?}", e);
-        }
-
         let (tx, rx) = mpsc::channel();
 
         if self.worker_outbox.send(WorkerMessage::Stop(tx)).is_err() {
@@ -506,7 +525,7 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
             .write(true)
             .open(path.join("log"))?;
 
-        let sstable_directory = list_sstables(path)?;
+        let sstable_directory = list_sstables(path, true)?;
 
         let mut db = BTreeMap::new();
         for sstable_id in sstable_directory.keys() {
@@ -573,12 +592,18 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
 
         let (tx, rx) = mpsc::channel();
 
+        let worker_stats = Arc::new(WorkerStats {
+            read_bytes: 0.into(),
+            written_bytes: 0.into(),
+        });
+
         let worker: Worker<K, V> = Worker {
             path: path.clone().into(),
             sstable_directory,
             inbox: rx,
             db_sz: db.len() as u64 * (K + V) as u64,
             config,
+            stats: worker_stats.clone(),
         };
 
         std::thread::spawn(move || worker.run());
@@ -588,13 +613,23 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
 
         let lsm = Lsm {
             log: BufWriter::with_capacity(config.log_bufwriter_size, reader.into_inner()),
-            memtable,
-            db,
             path: path.into(),
             next_sstable_id: max_sstable_id.unwrap_or(0) + 1,
             dirty_bytes: recovered as usize,
             worker_outbox: tx,
             config,
+            stats: Stats {
+                logged_bytes: recovered,
+                on_disk_bytes: 0,
+                read_bytes: 0,
+                written_bytes: 0,
+                resident_bytes: db.len() as u64 * (K + V) as u64,
+                space_amp: 0.,
+                write_amp: 0.,
+            },
+            worker_stats,
+            db,
+            memtable,
         };
 
         Ok(lsm)
@@ -627,19 +662,23 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
 
     fn log_mutation(&mut self, k: [u8; K], v: Option<[u8; V]>) -> Result<()> {
         let crc: u32 = hash(&k, &v);
-        self.log.write(&crc.to_le_bytes())?;
-        self.log.write(&[v.is_some() as u8])?;
-        self.log.write(&k)?;
+        self.log.write_all(&crc.to_le_bytes())?;
+        self.log.write_all(&[v.is_some() as u8])?;
+        self.log.write_all(&k)?;
 
         if let Some(v) = v {
-            self.log.write(&v)?;
+            self.log.write_all(&v)?;
         } else {
-            self.log.write(&[0; V])?;
+            self.log.write_all(&[0; V])?;
         };
 
+        let logged_bytes = 4 + 1 + K + V;
+
         self.memtable.insert(k, v);
 
-        self.dirty_bytes += 4 + K + V;
+        self.dirty_bytes += logged_bytes;
+        self.stats.logged_bytes += logged_bytes as u64;
+        self.stats.written_bytes += logged_bytes as u64;
 
         if self.dirty_bytes > self.config.max_log_length {
             self.flush()?;
@@ -661,26 +700,28 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
         self.log.get_mut().sync_all()?;
 
         if self.dirty_bytes > self.config.max_log_length {
-            log::debug!("flushing log");
+            log::debug!("compacting log to sstable");
             let memtable = std::mem::take(&mut self.memtable);
             let sst_id = self.next_sstable_id;
             if let Err(e) = write_sstable(&self.path, sst_id, &memtable, false, &self.config) {
                 // put memtable back together before returning
                 self.memtable = memtable;
-                log::error!("failed to flush lsm log to sst: {:?}", e);
+                log::error!("failed to flush lsm log to sstable: {:?}", e);
                 return Err(e.into());
             }
 
             let sst_sz = 8 + (memtable.len() as u64 * (4 + K + V) as u64);
             let db_sz = self.db.len() as u64 * (K + V) as u64;
 
-            self.worker_outbox
-                .send(WorkerMessage::NewSST {
-                    id: sst_id,
-                    sst_sz,
-                    db_sz,
-                })
-                .expect("compactor must outlive db");
+            if let Err(e) = self.worker_outbox.send(WorkerMessage::NewSST {
+                id: sst_id,
+                sst_sz,
+                db_sz,
+            }) {
+                log::error!("failed to send message to worker: {:?}", e);
+                log::logger().flush();
+                panic!("failed to send message to worker: {:?}", e);
+            }
 
             self.next_sstable_id += 1;
 
@@ -697,6 +738,27 @@ impl<const K: usize, const V: usize> Lsm<K, V> {
 
         Ok(())
     }
+
+    pub fn stats(&mut self) -> Result<Stats> {
+        self.stats.written_bytes += self.worker_stats.written_bytes.swap(0, Ordering::Relaxed);
+        self.stats.read_bytes += self.worker_stats.read_bytes.swap(0, Ordering::Relaxed);
+        self.stats.resident_bytes = self.db.len() as u64 * (K + V) as u64;
+
+        let mut on_disk_bytes: u64 = std::fs::metadata(self.path.join("log"))?.len();
+
+        on_disk_bytes += list_sstables(&self.path, false)?
+            .into_iter()
+            .map(|(_, len)| len)
+            .sum::<u64>();
+
+        self.stats.on_disk_bytes = on_disk_bytes;
+
+        self.stats.write_amp =
+            self.stats.written_bytes as f64 / self.stats.on_disk_bytes.max(1) as f64;
+        self.stats.space_amp =
+            self.stats.on_disk_bytes as f64 / self.stats.resident_bytes.max(1) as f64;
+        Ok(self.stats)
+    }
 }
 
 #[cfg(test)]
